
/******************************************************************************
* MODULE     : evaluate.cc
* DESCRIPTION: The main evaluation routines
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <evaluator.hh>
#include <single.hh>
#include <literal.hh>
#include <xtable.hh>
#include <keyword.hh>
#include <error.hh>
#include <function.hh>
namespace mmx {

GC_GLOBAL (bool, re_entrant_flag, false);

/******************************************************************************
* The main loop
******************************************************************************/

void
evaluator_rep::interrupt () {
#ifdef SUPPORTS_GC
  if (!re_entrant_flag) gc_collect ();
#endif
  interrupt_evaluation= false;
}

generic
evaluator_rep::loop () {
  while (!nil (cur)) {
    if (interrupt_evaluation) interrupt ();
    //cout << "Trace " << trace << lf;
    //if (!nil (trace)) cout << "Evaluate " << car(trace) << ", " << cur << lf;
    //cout << "Evaluate " << N(trace) << ", " << cur << lf;
    //cout << "Evaluate " << cur << lf;
    //cout << "Evaluate " << cur << ", " << rem << lf;
    //cout << "Environment " << env->local_level() << ", " <<N(env) << ", " << ((int) env.rep) << lf;

    switch (gentype (cur)) {
    case STRING_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (value (tmp, ID_STRING), cdr (rem));
      break;
    }
    case SYMBOL_TYPE:
      if (env->contains (cur)) {
	generic tmp= cur;
	cur= car (rem);
	rem= cons (env [tmp], cdr (rem));
	//cout << "  --> " << car (rem) << lf;
      }
      else {
	// cout << "env= " << env << lf;
	error (cur, "Unknown identifier");
      }
      break;
    case LIST_TYPE:
      trace= cons (cur, trace);
      rem  = cons (KW_COMPOUND_1, rem);
      cur  = car (cur);
      break;
    case SINGLE_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (value (tmp, ID_INT), cdr (rem));
      break;
    }
    case VALUE_TYPE: {
      generic tmp= cur;
      cur= car (rem);
      rem= cons (tmp, cdr (rem));
      break;
    }
    case LITERAL_INT_TYPE: {
      generic g= env [MMX_NEW_INT];
      generic v= value (cur, ID_LITERAL_INT);
      if (nil (g))
	error (cur, "No method for literal integers available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case LITERAL_FLOAT_TYPE: {
      generic g= env [MMX_NEW_FLOAT];
      generic v= value (cur, ID_LITERAL_FLOAT);
      if (nil (g))
	error (cur, "No method for literal floats available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case LITERAL_STRING_TYPE: {
      generic g= env [MMX_NEW_STRING];
      generic v= value (cur, ID_LITERAL_STRING);
      if (nil (g))
	error (cur, "No method for literal strings available");
      else {
	trace= cons (cur, trace);
	cur= KW_COMPOUND_2;
	rem= cons (v, list (g), list (), rem);
	kw_compound_2 ();
      }
      break;
    }
    case KEYWORD_TYPE:
      //cout << "Evaluate " << cur << lf;
      (this ->* (as_keyword (cur)->fn)) ();
      break;
    default:
      error (cur, "Syntax error");
    }
  }
  //cout << "env= " << env << lf;
  return car (rem);
}

/******************************************************************************
* Evaluating expressions
******************************************************************************/

generic
evaluator_rep::evaluate (generic g) {
  GC_LOCAL (generic, arg, g);
  GC_LOCAL (generic, ret, generic ());

  env (MISC_TODO)= list ();
  env (MISC_DONE)= list ();
  trace= list ();
  cur= arg;
  rem= list (generic ());
  ret= loop ();
  /*
#ifdef DEBUG_ON
  while (env->local_level () != 0) {
    xtable t= as_xtable (env);
    cout << "Pending local variables" << lf << hrule;
    dump (t->ext_t);
    cout << hrule;
    env= t->base_t;
  }
#endif
  */
  return ret;
}

void
evaluator_rep::error (generic where, string message) {
#ifdef DEBUG_ON
  cout << hrule;
  cout << "cont = " << cons (cur, rem) << lf;
  cout << hrule;
  cout << "trace= " << trace << lf;
  cout << hrule;
#endif
  generic err= std_error (message, where);
  cur= generic ();
  rem= cons (list (symbol ("errors!"), err), rem);
}

/******************************************************************************
* Extensions
******************************************************************************/

generic
extern_eval (generic g) {
  // cout << "g= " << g << lf;
  GC_LOCAL (generic , r         , generic ());
  GC_LOCAL (gentable, save_env  , mmeval->env);
  GC_LOCAL (generic , save_cur  , mmeval->cur);
  GC_LOCAL (list    , save_rem  , mmeval->rem);
  GC_LOCAL (list    , save_trace, mmeval->trace);
  GC_LOCAL (bool    , save_flag , re_entrant_flag);
  re_entrant_flag= true;
  mmeval->start_local_env (table ());
  r= mmx_evaluate (g);
  mmeval->env    = save_env;
  mmeval->cur    = save_cur;
  mmeval->rem    = save_rem;
  mmeval->trace  = save_trace;
  re_entrant_flag= save_flag;
  // cout << "r= " << r << lf;

  if (is_value (r)) return as_value (r) -> val;
  if (is_func (r, symbol ("errors!"))) {
    list errs= cdr (as_list (r));
    print_errors (errs);
  }
  fatal_error ("Errors in re-entrant code are fatal",
	       "extern_eval", "evaluate.cc");
  return generic ();
}

bool
extern_equal_types (generic t1, generic t2) {
  // FIXME: how to safely test for type expressions?
  // We now use the following dirty hack so that template arguments 1#, 2#, ...
  // match any type (which of course may not be the case).
  if (is_symbol (t1) && is_string (S (t1)) && ends (as_string (S (t1)), "#"))
    return true;
  if (is_symbol (t2) && is_string (S (t2)) && ends (as_string (S (t2)), "#"))
    return true;
  // End dirty hack

  if (nil (t1) || nil (t2)) return nil (t1) && nil (t2);
  else if (is_mirror (t1) && is_mirror (t2)) return t1 == t2;
  else if (is_list (t1) && is_list (t2))
    return extern_equal_types (car (t1), car (t2)) &&
           extern_equal_types (cdr (t1), cdr (t2));
  else return false;
}

/******************************************************************************
* The user interface
******************************************************************************/

generic
mmx_evaluate (generic g) {
  return mmeval->evaluate (g);
}

void
mmx_display (value v) {
  return mmeval->display (v);
}

} // namespace mmx
