
/******************************************************************************
* MODULE     : glued.cc
* DESCRIPTION: Externally glued objects
* COPYRIGHT  : (C) 2004  Bernard Mourrain and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_GLUED_HH
#define __MMX_GLUED_HH
#include <config.hh>
#include <generic.hh>
#include <array.hh>
#include <list.hh>
#include <string.hh>
#include <gentable.hh>
#include <table.hh>
#include <symbol.hh>
#include <single.hh>
#include <printer.hh>
#include <literal.hh>
#include <alias.hh>
#include <mirror.hh>
#include <value.hh>
#include <function.hh>
#include <symbols.hh>
namespace mmx {

/******************************************************************************
* Registration and re-entrancy
******************************************************************************/

void register_initializer (string prefix, list files, function init);
void declare (symbol var_name, generic val, generic type);
void declare (symbol var_name, mmx_function fn, generic type);
void call_initializer (string prefix, string file);
generic extern_eval (generic g);
bool extern_equal_types (generic t1, generic t2);
// defined in extern.cc and evaluate.cc

class auto_initializer {
public:
  inline auto_initializer (string prefix, list files, function init) {
    register_initializer (prefix, files, init);
  }
};

/******************************************************************************
* The glued template class
******************************************************************************/

template<typename T, typename S> class glued;

template<typename T, typename S>
class glued_rep: public generic_rep {
public:
  T data;           // the glued data
  S type;           // type of the data
  bool alloc_flag;  // false if reference

protected:
  size_t Size () { return sizeof (glued_rep<T,S>); }
  void Mark () {
    // cannot do better for the moment for data
    mark (type);
  }
  int Type () { return GLUED_TYPE; }
  void Print (ostream& out) {
    out << "#glued<";
    type->gen_lessless (out, data);
    out << " ";
    out << type->mmxt;
    out << ">";
  }
  int Hash () { return type->gen_hash (data); }
  bool Equal (generic g);
  generic Copy ();
  void Assign_me (generic g);

public:
  inline glued_rep (T data2, S type2, bool alloc_flag2):
    data (data2), type (type2), alloc_flag (alloc_flag2) {}
  inline ~glued_rep () { if (alloc_flag) type->gen_delete (data); }
  inline T copy_data () { return type->gen_new (data); }

  friend class glued<T,S>;
};

template<typename T, typename S>
class glued {
DEFAULT_METHODS_PROTO_2(glued,T,S);
public:
  inline glued<T,S> (T data, S type, bool alloc_flag= true):
    rep (new glued_rep<T,S> (data, type, alloc_flag)) {}
};
DEFAULT_METHODS_CODE_2(glued,T,S);

template<typename T, typename S> bool
glued_rep<T,S>::Equal (generic g) {
  if (!is_glued (g)) return false;
  glued<T,S> gl= as_glued<T,S> (g);
  if (!extern_equal_types (type->mmxt, gl->type->mmxt)) return false;
  return type->gen_equal (data, gl->data);
}

template<typename T, typename S> generic
glued_rep<T,S>::Copy () {
  return glued<T,S> (copy_data (), type);
}

template<typename T, typename S> void
glued_rep<T,S>::Assign_me (generic g) {
  assert (is_glued (g));
  glued<T,S> gl= as_glued<T,S> (g);
  if (!extern_equal_types (type->mmxt, gl->type->mmxt)) {
    cerr << "Assignment " << type->mmxt << " <- " << gl->type->mmxt << "\n";
    assert (extern_equal_types (type->mmxt, gl->type->mmxt));
  }
  type->gen_assign (data, gl->data);
}

/******************************************************************************
* Main converters between glue and generics
******************************************************************************/

template<typename C, typename T, typename S>
class exporter {
public:
  static inline generic
  helper (const C& x, const S& type) {
    return glued<T,S> (T ((void*) new C(x)), type);
  }
};

template<typename C, typename T, typename S>
class importer {
public:
  static inline C
  helper (const generic& g) {
    return *((C*) as_glued<T,S>(g)->data.rep);
  }
};

template<typename C, typename T, typename S> generic
as_mmx (const C& x, const S& type) {
  return exporter<C, T, S>::helper (x, type);
}

template<typename C, typename T, typename S> C
as_cpp (const generic& g) {
  return importer<C, T, S>::helper (g);
}

/******************************************************************************
* Converters for scalar built-in Mathemagix types
******************************************************************************/

template<typename T, typename S>
class exporter<int, T, S> {
public:
  static inline generic
  helper (const int& x, const S& type) {
    return single (x);
  }
};

template<typename T, typename S>
class importer<int, T, S> {
public:
  static inline int
  helper (const generic& g) {
    return as_int (g);
  }
};

template<typename T, typename S>
class exporter<bool, T, S> {
public:
  static inline generic
  helper (const bool& x, const S& type) {
    return single (x);
  }
};

template<typename T, typename S>
class importer<bool, T, S> {
public:
  static inline bool
  helper (const generic& g) {
    return as_bool (g);
  }
};

template<typename T, typename S>
class exporter<string, T, S> {
public:
  static inline generic
  helper (const string& s, const S& type) {
    return s;
  }
};

template<typename T, typename S>
class importer<string, T, S> {
public:
  static inline string
  helper (const generic& g) {
    return as_string (g);
  }
};

template<typename T, typename S> generic
as_mmx_ostream (ostream& x, const S& type) {
  return printer (x);
}

template<typename T, typename S> ostream&
as_cpp_ostream (const generic& g) {
  return as_ostream (g);
}

/******************************************************************************
* Aliases
******************************************************************************/

template<typename C>
class xalias_rep: public generic_rep {
protected:
  void Print (ostream& out) { out << "#xalias<" << Read() << ">"; }
  int Type () { return XALIAS_TYPE; }

public:
  virtual C Read () = 0;
  virtual C& Write () = 0;
};

template<typename C>
class std_xalias_rep: public xalias_rep<C> {
public:
  C&   ref;         // the reference
  bool alloc_flag;  // destroy &ref on deletion?

protected:
  size_t Size () { return sizeof (std_xalias_rep<C>); }
  void Mark () {
    // FIXME: cannot do better for the moment for ref
  }

public:
  C Read () { return ref; }
  C& Write () { return ref; }
  inline std_xalias_rep (C& c, bool f): ref (c), alloc_flag (f) {}
  inline ~std_xalias_rep () { if (alloc_flag) delete &ref; }
};

template<typename C>
class xalias {
DEFAULT_METHODS_PROTO_1(xalias,C);
public:
  inline xalias (C& c, bool f= false):
    rep (new std_xalias_rep<C> (c, f)) {}
  operator C () const { return rep->Read (); }
  operator C& () const { return rep->Write (); }
};
DEFAULT_METHODS_CODE_1(xalias,C);

template<typename C, typename T, typename S>
class import_xalias_rep: public xalias_rep<C> {
public:
  genalias ref;

protected:
  size_t Size () { return sizeof (import_xalias_rep<C,T,S>); }
  void Mark () { mark (ref); }

public:
  C Read () { return as_cpp<C,T,S> (ref->get ()); }
  C& Write () {
    generic& g= ref->get_ref ();
    return *((C*) as_glued<T,S>(g)->data.rep);
  }
  inline import_xalias_rep (genalias g): ref (g) {}
};

template<typename C, typename T, typename S>
class importer<xalias<C>, T, S> {
public:
  static inline xalias<C>
  helper (const generic& g) {
    assert (is_genalias (g));
    return new import_xalias_rep<C,T,S> (as_genalias (g));
  }
};

template<typename C, typename T, typename S>
class exporter<xalias<C>, T, S> {
public:
  static inline generic
  helper (const xalias<C>& ref, const S& type) {
    // NOTE: type contains scalar type here
    C& the_ref= (C&) ref;
    T data ((void*) &the_ref);
    generic* ptr= (generic*) new glued<T,S> (data, type, false);
    return alias (*ptr, generic (), true);
  }
};

/******************************************************************************
* Tuples and vtuples
******************************************************************************/

template<typename C>
class xtuple_rep: public generic_rep {
protected:
  int Type () { return XTUPLE_TYPE; }

public:
  virtual int Length () = 0;
  virtual C Element (int i) = 0;
};

template<typename C>
class std_xtuple_rep: public xtuple_rep<C> {
public:
  C*  a;
  int n;

protected:
  size_t Size () { return sizeof (std_xtuple_rep<C>); }
  void Mark () {
    // FIXME: cannot do better for the moment for a
    mark (n);
  }
  int Length () { return n; }
  C Element (int i) { return a[i]; }

public:
  std_xtuple_rep (C* a2, int n2): a (NULL), n (n2) {
    if (n != 0) {
      a= new C[n];
      for (int i=0; i<n; i++)
	a[i]= a2[i];
    }
  }
  ~std_xtuple_rep () {
    if (n != 0) delete[] a;
  }
};

template<typename C>
class xtuple {
DEFAULT_METHODS_PROTO_1(xtuple,C);
public:
  inline xtuple (C* a, int n): rep (new std_xtuple_rep<C> (a, n)) {}
  inline C operator [] (int i) const { return rep->Element (i); }
  inline int length () const { return rep->Length (); }
};
DEFAULT_METHODS_CODE_1(xtuple,C);

template<typename C> int
N (const xtuple<C>& t) {
  return t.length ();
}

template<typename C, typename T, typename S>
class import_xtuple_rep: public xtuple_rep<C> {
public:
  array a;

protected:
  size_t Size () { return sizeof (import_xtuple_rep<C,T,S>); }
  void Mark () { mark (a); }
  void Print (ostream& out) { out << "#import_xtuple<" << a << ">"; }

public:
  int Length () { return N(a); }
  C Element (int i) { return as_cpp<C,T,S> (a[i]); }
  inline import_xtuple_rep (array a2): a (a2) {}
};

static generic
decode_tuple (generic g) {
  if (is_func (g, MMX_TUPLE) || is_func (g, MMX_VTUPLE)) {
    int  i= 0;
    list l= cdr (as_list (g));
    array a (N(l));
    for (i=0; !nil (l); i++, l= cdr (l))
      a[i]= decode_tuple (car (l));
    return a;
  }
  else return g;
}

template<typename C, typename T, typename S>
class importer<xtuple<C>, T, S> {
public:
  static inline xtuple<C>
  helper (const generic& g) {
    assert (is_func (g, MMX_TUPLE) || is_func (g, MMX_VTUPLE) || is_array (g));
    return new import_xtuple_rep<C,T,S> (as_array (decode_tuple (g)));
  }
};

static generic
encode_tuple (array a) {
  int i, n= N(a);
  list l;
  for (i=n-1; i>=0; i--)
    l= cons (a[i], l);
  return cons (MMX_TUPLE, l);
}

template<typename C, typename T, typename S>
class exporter<xtuple<C>, T, S> {
public:
  static inline generic
  helper (const xtuple<C>& xt, const S& type) {
    // NOTE: type contains scalar type here
    int i, n= N(xt);
    array a (n);
    for (i=0; i<n; i++)
      a[i]= as_mmx<C,T,S> (xt[i], type);
    return encode_tuple (a);
  }
};

/******************************************************************************
* Conversions from and to gentype structures
******************************************************************************/

template<typename S>
class type_converter_rep: public generic_rep {
public:
  table conv_routine;
  table conv_cache;

protected:
  inline size_t Size () { return sizeof (type_converter_rep<S>); }
  inline void Mark () { mark (conv_routine); mark (conv_cache); }
  inline int Type () { return TYPE_CONVERTER_TYPE; }
  inline void Print (ostream& out) { out << "#type_converter<>"; }

public:
  inline type_converter_rep () {}
  inline void declare (symbol name, generic type_id, mmx_function fn);
  inline generic convert (generic g);
};

template<typename S>
class type_converter {
DEFAULT_METHODS_PROTO_1(type_converter,S);
public:
  inline type_converter (): rep (new type_converter_rep<S> ()) {}
};
DEFAULT_METHODS_CODE_1(type_converter,S);

template<typename S> GC_REGISTER_TYPE (type_converter<S> );
#define TMPL_GC_REGISTER_TYPE(type) template<> GC_REGISTER_TYPE(type)

template<typename S> void
type_converter_rep<S>::declare (symbol name, generic id, mmx_function fn) {
  conv_routine (id)= function (name, fn);
}

template<typename S> generic
type_converter_rep<S>::convert (generic g) {
  if (conv_cache->contains (g)) return conv_cache [g];
  generic h= (is_list (g)? g: generic (list (g))), ret;
  if (!conv_routine->contains (car (h))) ret= S (g);
  else ret= as_function (conv_routine (car (h))) (g);
  conv_cache (g)= ret;
  return ret;
}

#define GENTYPE_CONVERTERS(CV,C,R)		\
namespace mmx {					\
  template<typename T, typename S>		\
  class exporter<C, T, S> {			\
  public:					\
    static inline generic			\
    helper (const C& x, const S& type) {	\
      return x->mmxt;				\
    }						\
  };						\
						\
  template<typename T, typename S>		\
  class importer<C, T, S> {			\
  public:					\
    static C					\
    helper (const generic& g) {			\
      return as_##C (CV->convert (g));		\
    }						\
  };						\
}

/******************************************************************************
* Re-entrant objects and conversions from and to geninst structures
******************************************************************************/

#define MMX_INC_COUNT(x) { (x)->ref_count++; }
#define MMX_DEC_COUNT(x) { if (0==--((x)->ref_count)) { delete (x); } }

class external_rep {
MMX_ALLOCATORS;
public:
  int ref_count;
  generic g;

public:
  inline external_rep (generic g2): ref_count (1), g (g2) {
    gc_register ("glued", &g); }
  inline ~external_rep () {
    gc_unregister ("glued", &g); }
  friend class external;
};

class external {
MMX_ALLOCATORS;
public:
  external_rep* rep;

public:
  inline external (const external& x): rep(x.rep) {
    MMX_INC_COUNT (rep); }
  inline ~external () {
    MMX_DEC_COUNT (rep); }
  inline external& operator = (const external& x) {
    MMX_INC_COUNT (x.rep); MMX_DEC_COUNT (rep);
    rep=x.rep; return *this; }
  inline external (generic g): rep (new external_rep (g)) {}
  inline operator generic () const { return rep->g; }
  inline operator generic& () const { return rep->g; }
};

#undef MMX_INC_COUNT
#undef MMX_DEC_COUNT

template<typename T, typename S>
class exporter<T, T, S> {
public:
  static inline generic
  helper (const T& x, const S& type) {
    return (generic) *((external*) x.rep);
  }
};

template<typename T, typename S>
class importer<T, T, S> {
public:
  static inline T
  helper (const generic& g) {
    return T ((void*) new external (g));
  }
};

/******************************************************************************
* Conversions from and to geninst_nr structures
******************************************************************************/

#define GENINSTX_CONVERTERS(C,X)					\
namespace mmx {								\
  template<typename T, typename S>					\
  class exporter<C, T, S> {						\
  public:								\
    static generic							\
    helper (const C& x, const S& type) {				\
      if (X->internal())						\
        return glued<T,S> (type->gen_new (T(x.rep)), type);		\
      else return (generic) *((external*) x.rep);			\
    }									\
  };									\
									\
  template<typename T, typename S>					\
  class importer<C, T, S> {						\
  public:								\
    static C								\
    helper (const generic& g) {						\
      if (X->internal()) return C (as_glued<T,S>(g)->copy_data().rep);	\
      else return C ((void*) new external (g));				\
    }									\
  };									\
									\
  template<typename T, typename S>					\
  class exporter<xalias<C>, T, S> {					\
  public:								\
    static inline generic						\
    helper (const xalias<C>& ref, const S& type) {			\
      /* NOTE: type contains scalar type here */			\
      C& x= (C&) ref;							\
      if (X->internal()) {						\
	generic* ptr= (generic*) new glued<T,S> (x.rep, type, false);	\
	return alias (*ptr, generic (), true);				\
      }									\
      else return alias ((generic&) *((external*) x.rep), generic (), false); \
    }									\
  };									\
									\
  template<typename T, typename S>					\
  class importer<xalias<C>, T, S> {					\
  public:								\
    static xalias<C>							\
    helper (const generic& g) {						\
      generic& ref= as_genalias (g) -> get_ref ();			\
      if (X->internal()) {						\
        C* ptr= (C*) ((void*) &as_glued<T,S> (ref) -> data);		\
        return xalias<C> (*ptr, false);					\
      }									\
      else {								\
        C* ptr= new C ((void*) (&ref));					\
        return xalias<C> (*ptr, true);					\
      }									\
    }									\
  };									\
}

} //namespace mmx

/******************************************************************************
* Some default templates
******************************************************************************/

template<typename T> std::ostream&
print_lisp (std::ostream& out, T x) {
  return out << x;
}

template<typename T> T
div (T x, T y) {
  return x/y;
}

template<typename T> T
mod (T x, T y) {
  return x%y;
}

template<class T, class F> inline T
cast (const F& x) {
  return *reinterpret_cast<T*> (const_cast<F*> (&x));
}

#endif // __MMX_GLUED_HH
