
/******************************************************************************
* MODULE     : pretty.cc
* DESCRIPTION: Pretty printer for mathemagix trees
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <symbols.hh>
#include <list.hh>
#include <literal.hh>
#include <single.hh>
namespace mmx {

 
#define DATA_BEGIN   ((char) 2)
#define DATA_END     ((char) 5)
#define DATA_ESCAPE  ((char) 27)

GC_GLOBAL (bool, texmacs_mode, false);
GC_GLOBAL (bool, brown_flag, false);

GC_GLOBAL_STATIC (table, op_type, table ());
GC_GLOBAL_STATIC (table, op_name, table ());
GC_GLOBAL_STATIC (table, op_tmname, table ());
GC_GLOBAL_STATIC (table, special_symbol, table ());

/******************************************************************************
* The pretty printer class
******************************************************************************/

class pretty_printer_rep: public generic_rep {
private:
  ostream& out;
  list     sl;
  int      tabs;

protected:
  size_t Size ();
  void Mark ();
  int Type ();
  void Print (ostream& out);

private:
  void pp_spc ();
  void pp_lf ();
  void pp_indent ();
  void pp_unindent ();

  void pp_sep (bool large);
  void pp_open ();
  void pp_close (bool large);
  void pp_op (generic op, bool infix= true);
  void pp_verb (char* s1, char* s2);
  void pp_text (generic g);
  void pp_keyw (bool before, symbol s, bool after);

  void pp_C0 (generic g, bool large= false);
  void pp_Conds (generic g);
  void pp_Splitl (generic g);
  void pp_Casel (generic g);
  void pp_Among (generic g);
  void pp_Cl (generic g);
 
  void pp_E0 (generic g);
  void pp_E1 (generic g);
  void pp_Lflux (generic g);
  void pp_Rflux (generic g);

  void pp_E0s (generic g);
  void pp_Vt (generic g);
  void pp_Matrix (generic g);
  void pp_Vtuple (generic g);

  void pp_F0 (generic g);
  void pp_Seqors (generic g);
  void pp_Ors (generic g);
  void pp_Xors (generic g);
  void pp_F1 (generic g);
  void pp_Seqands (generic g);
  void pp_Ands (generic g);
  void pp_F2 (generic g);
  void pp_F3 (generic g);

  void pp_T0 (generic g);
  void pp_T1 (generic g);
  void pp_T2 (generic g);
  void pp_T3 (generic g);
  void pp_T4 (generic g);
  void pp_T5 (generic g);
  void pp_DomIm (generic g, bool flag);

  void pp_R0 (generic g);
  void pp_R1 (generic g);
  void pp_R2 (generic g);
  void pp_R3 (generic g);

public:
  inline pretty_printer_rep (ostream& out2): out (out2), tabs (0) {}
  inline ~pretty_printer_rep () {}

  void pp (generic g);

  friend class pretty_printer;
};

class pretty_printer {
DEFAULT_METHODS(pretty_printer);
public:
  inline pretty_printer (ostream& out):
    rep (new pretty_printer_rep (out)) {}
};

/******************************************************************************
* Required routines for generic structures
******************************************************************************/

size_t
pretty_printer_rep::Size () {
  return sizeof (pretty_printer_rep);
}

void
pretty_printer_rep::Mark () {
  mark (out);
  mark (sl);
  mark (tabs);
}

int
pretty_printer_rep::Type () {
  return PRETTY_PRINTER_TYPE;
}

void
pretty_printer_rep::Print (ostream& out) {
  out << "#pretty_printer<>";
}

/******************************************************************************
* Operator names
******************************************************************************/

static void
add (symbol var, symbol type, symbol val, symbol tmval) {
  op_type   (var)= type;
  op_name   (var)= val;
  op_tmname (var)= tmval;
}

static void
init () {
  if (N(op_type) == 0) {
    add (MMX_NUMBER, "operator", "#", "\\#");
    add (MMX_NOT, "operator", "!", "\\neg");
    add (MMX_PREINC, "operator", "++", "++");
    add (MMX_PREDEC, "operator", "--", "--");
    add (MMX_PRIME, "operator", "'", "'");
    add (MMX_BACKPRIME, "operator", "`", "^{\\backprime}");
    add (MMX_FACTORIAL, "postfix", "!", "!");
    add (MMX_POSTINC, "postfix",  "++", "++");
    add (MMX_POSTDEC, "postfix", "--", "--");
    add (MMX_DEFINE, "operator", "==", "==");
    add (MMX_DEFINEALIAS, "operator", ":=", "\\assign");
    add (MMX_ASSIGN, "operator", ":=", "\\assign");
    add (MMX_PLUSASSIGN, "operator", "+=", "\\plusassign");
    add (MMX_MINUSASSIGN, "operator", "-=", "\\minusassign");
    add (MMX_TIMESASSIGN, "operator", "*=", "*=");
    add (MMX_OVERASSIGN, "operator", "/=", "/=");
    add (MMX_LESSLESS, "operator", "<<", "\\ll");
    add (MMX_LESSLESS_VAR, "operator", "<<*", "\\ll_*");
    add (MMX_LESSLESS_STR, "operator", "<<%", "\\ll_\\%");
    add (MMX_GTRGTR, "operator", ">>", "\\gg");
    add (MMX_SEQOR, "operator", "or", "\\text{or}");
    add (MMX_OR, "operator", "||", "\vee");
    add (MMX_XOR, "operator", "^^", "\\triangledown");
    add (MMX_SEQAND, "operator", "and", "\\text{and}");
    add (MMX_AND, "operator", "&&", "\\wedge");
    add (MMX_IMPLIES, "operator", "=>", "\\Rightarrow");
    add (MMX_EQUIV, "operator", "<=>", "\\Leftrightarrow");
    add (MMX_EQUAL, "operator", "=", "=");
    add (MMX_LESS, "operator", "<", "<");
    add (MMX_LESSEQ, "operator", "<=", "\\leqslant");
    add (MMX_GTR, "operator", ">", ">");
    add (MMX_GTREQ, "operator", ">=", "\\geqslant");
    add (MMX_UNEQUAL, "operator", "!=", "\\neq");
    add (MMX_NOTLESS, "operator", "!<", "\\nless");
    add (MMX_NOTLESSEQ, "operator", "!<=", "\\nleqslant");
    add (MMX_NOTGTR, "operator", "!>", "\\ngtr");
    add (MMX_NOTGTREQ, "operator", "!>=", "\\ngtrslant");
    add (MMX_TYPE, "operator", ":", ":");
    add (MMX_TRANSTYPE, "operator", ":>", "\\;:\\!\\!\\!>");
    add (MMX_FUNCTION, "operator", "->", "\\rightarrow");
    add (MMX_DOTS, "operator", "..", "\\ldots");
    add (MMX_PLUS, "operator", "+", "+");
    add (MMX_OPLUS, "operator", "@+", "\\oplus");
    add (MMX_MINUS, "operator", "-", "-");
    add (MMX_OMINUS, "operator", "@-", "\\ominus");
    add (MMX_TIMES, "operator", "*", "\\*");
    add (MMX_OTIMES, "operator", "@*", "\\otimes");
    add (MMX_OVER, "operator", "/", "/");
    add (MMX_OOVER, "operator", "@/", "\\oslash");
    add (MMX_COMPOSE, "operator", "@", "\\circ");
    add (MMX_POWER, "operator", "^", "\\hat{}");
  }
}

symbol
mmx_op_type (generic s) {
  init ();
  if (op_type->contains (s)) return as_symbol (op_type[s]);
  else return symbol ("unknown");
}

symbol
mmx_op_name (generic s) {
  init ();
  if (op_name->contains (s)) return as_symbol (op_name[s]);
  else return as_symbol (s);
}

symbol
mmx_op_tmname (generic s) {
  init ();
  if (op_tmname->contains (s)) return as_symbol (op_tmname[s]);
  else return as_symbol (s);
}

/******************************************************************************
* Special symbols
******************************************************************************/

void
add_special (string s) {
  special_symbol (s)= single (true);
}

void
add_greek (string s) {
  add_special (s);
  add_special (string ((char) ((int) s[0] - 32)) * s (1, N(s)));
}

void
init_special () {
  add_special ("mathd");
  add_special ("mathe");
  add_special ("mathi");
  add_special ("mathpi");
  add_special ("partial");
  add_greek ("alpha");
  add_greek ("beta");
  add_greek ("gamma");
  add_greek ("delta");
  add_greek ("epsilon");
  add_greek ("zeta");
  add_greek ("eta");
  add_greek ("theta");
  add_greek ("iota");
  add_greek ("kappa");
  add_greek ("lambda");
  add_greek ("mu");
  add_greek ("nu");
  add_greek ("xi");
  add_greek ("omicron");
  add_greek ("pi");
  add_greek ("rho");
  add_greek ("sigma");
  add_greek ("tau");
  add_greek ("upsilon");
  add_greek ("phi");
  add_greek ("psi");
  add_greek ("chi");
  add_greek ("omega");
}

bool
is_special_symbol (generic s) {
  if (N (special_symbol) == 0) init_special ();
  return special_symbol->contains (s);
}

/******************************************************************************
* Subroutines
******************************************************************************/

void
pretty_printer_rep::pp_spc () {
  if (texmacs_mode) out << "\\;";
  else out << " ";
}

void
pretty_printer_rep::pp_lf () {
  if (texmacs_mode) {
    int i;
    out << "\\\\";
    for (i=0; i<tabs; i++)
      out << "\\;\\;\\;\\;";
  }
  else out << lf;
}

void
pretty_printer_rep::pp_indent () {
  if (texmacs_mode) {
    tabs++;
    out << "\\;\\;\\;\\;";
  }
  else out << indent;
}

void
pretty_printer_rep::pp_unindent () {
  if (texmacs_mode) {
    tabs--;
    out << "\b\b\b\b\b\b\b\b";
  }
  else out << unindent;
}

void
pretty_printer_rep::pp_sep (bool large) {
  if (large) {
    out << ";";
    pp_lf ();
  }
}

void
pretty_printer_rep::pp_open () {
  if (texmacs_mode) out << "\\{";
  else out << "{";
  pp_lf ();
  pp_indent ();
}

void
pretty_printer_rep::pp_close (bool large) {
  pp_unindent ();
  if (texmacs_mode) out << "\\}";
  else out << "}";
  if (large) pp_lf ();
}

void
pretty_printer_rep::pp_op (generic op, bool infix) {
  symbol s= as_symbol (op);
  if (texmacs_mode) out << mmx_op_tmname (s);
  else {
    if (infix) pp_spc ();
    out << mmx_op_name (s);
    if (infix) pp_spc ();
  }
}

void
pretty_printer_rep::pp_verb (char* s1, char* s2) {
  if (texmacs_mode) out << s1;
  else out << s2;
}

void
pretty_printer_rep::pp_text (generic g) {
  if (texmacs_mode) out << "\\text{";
  if (is_symbol (g) && is_string (S (as_symbol (g)))) {
    string s= as_string (S (as_symbol (g)));
    for (int i=0; i<N(s); i++)
      if (s[i] == '_' && texmacs_mode) out << "\\_";
      else out << s[i];
  }
  else out << g;
  if (texmacs_mode) out << "}";
}

void
pretty_printer_rep::pp_keyw (bool before, symbol s, bool after) {
  if (before) pp_spc ();
  if (texmacs_mode) out << "\\text{\\bf " << s << "}";
  else out << s;
  if (after) pp_spc ();
}

/******************************************************************************
* Predicates
******************************************************************************/

static bool
is_C0 (generic g) {
  if (is_func (g, MMX_BEGIN) ||
      is_func (g, MMX_LOCAL) ||
      is_func (g, MMX_IF) ||
      is_func (g, MMX_GENERATE) ||
      is_func (g, MMX_CASE) ||
      is_func (g, MMX_LOOP) ||
      is_func (g, MMX_SPLIT) ||
      is_func (g, MMX_RETURN) ||
      is_func (g, MMX_YIELD) ||
      is_func (g, MMX_BREAK) ||
      is_func (g, MMX_CONTINUE))
    return true;
  if ((N(g) != 3) || (!is_C0 (g[2]))) return false;
  return
    is_func (g, MMX_FORALL) ||
    is_func (g, MMX_EXISTS) ||
    is_func (g, MMX_OBTAIN) ||
    is_func (g, MMX_LAMBDA) ||
    is_func (g, MMX_DEFINE) ||
    is_func (g, MMX_ASSIGN) ||
    is_func (g, MMX_DEFINEALIAS) ||
    is_func (g, MMX_PLUSASSIGN) ||
    is_func (g, MMX_MINUSASSIGN) ||
    is_func (g, MMX_TIMESASSIGN) ||
    is_func (g, MMX_OVERASSIGN);
}

static bool
is_applicable (generic g) {
  if (N(g)>0) return true;
  if (is_literal (g)) return true;
  if (gentype (g) == SYMBOL_TYPE)
    return is_applicable (S(g));
  if (gentype (g) == STRING_TYPE) {
    string s= as_string (g);
    if (N(s)==0) return false;
    int i;
    for (i=0; i<N(s); i++)
      if (((s[i]<'0') || (s[i]>'9')) &&
          ((s[i]<'a') || (s[i]>'z')) &&
          ((s[i]<'A') || (s[i]>'Z')) &&
          ((s[i]<'_') || (s[i]>'_')))
        return false;
    return true;
  }
  return false;
}

/******************************************************************************
* Commands
******************************************************************************/

void
pretty_printer_rep::pp_C0 (generic g, bool large) {
  if (is_func (g, MMX_BEGIN)) {
    pp_open ();
    pp_Cl (cdr (g));
    pp_close (large);
  }
  else if (is_func (g, MMX_LOCAL)) {
    list l= cDr (cdr (as_list (g)));
    generic last= cAr (as_list (g));
    if (is_func (last, MMX_BEGIN)) l= l * cdr (as_list (last));
    else l= l * list (last);
    pp_C0 (cons (MMX_BEGIN, l), large);
  }
  else if (is_func (g, MMX_IF, 2) && large) {
    pp_keyw (false, "if", true);
    pp_E0 (g[1]);
    pp_keyw (true, "then", true);
    if (is_func (g[2], MMX_IF)) pp_C0 (list (MMX_BEGIN, g[2]), large);
    else pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_IF, 3)) {
    pp_keyw (false, "if", true);
    pp_E0 (g[1]);
    pp_keyw (true, "then", true);
    if (is_func (g[2], MMX_IF, 2)) pp_C0 (list (MMX_BEGIN, g[2]), large);
    else pp_C0 (g[2], large);
    pp_keyw (!large, "else", true);
    pp_C0 (g[3], large);
  }
  else if (is_func (g, MMX_GENERATE, 2)) {
    pp_keyw (false, "generate", true);
    pp_E0 (g[1]);
    pp_keyw (true, "do", true);
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_CASE, 2)) {
    pp_keyw (false, "case", true);
    pp_E0 (g[1]);
    pp_keyw (true, "from", true);
    pp_open ();
    pp_Casel (g[2]);
    pp_close (large);
  }
  else if (is_func (g, MMX_LOOP, 2)) {
    pp_Conds (g[1]);
    pp_keyw (false, "loop", true);
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_SPLIT)) {
    pp_keyw (false, "split", true);
    if (N(g) == 1) {
      pp_verb ("\\{\\}", "{}");
      if (large) {
	if (texmacs_mode);
	else out << lf;
      }
    }
    else {
      pp_open ();
      pp_Splitl (cdr (g));
      pp_close (large);
    }
  }
  else if (is_func (g, MMX_RETURN, 1)) {
    pp_keyw (false, "return", true);
    pp_C0 (g[1], large);
  }
  else if (is_func (g, MMX_YIELD, 1)) {
    pp_keyw (false, "yield", true);
    pp_C0 (g[1], large);
  }
  else if (g == MMX_RETURN) {
    pp_keyw (false, "return", false);
    pp_sep (large);
  }
  else if (g == MMX_YIELD) {
    pp_keyw (false, "yield", false);
    pp_sep (large);
  }
  else if (g == MMX_BREAK) {
    pp_keyw (false, "break", false);
    pp_sep (large);
  }
  else if (g == MMX_CONTINUE) {
    pp_keyw (false, "continue", false);
    pp_sep (large);
  }
  else if (is_func (g, MMX_FORALL, 2)) {
    pp_verb ("\\forall\\;", "forall (");
    pp_E0 (g[1]);
    pp_verb (",", ") ");
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_EXISTS, 2)) {
    pp_verb ("\\exists\\;", "exists (");
    pp_E0 (g[1]);
    pp_verb (",", ") ");
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_DEFINE, 2) ||
	   is_func (g, MMX_ASSIGN, 2) ||
	   is_func (g, MMX_DEFINEALIAS, 2) ||
	   is_func (g, MMX_PLUSASSIGN, 2) ||
	   is_func (g, MMX_MINUSASSIGN, 2) ||
	   is_func (g, MMX_TIMESASSIGN, 2) ||
	   is_func (g, MMX_OVERASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op (car (g));
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_OBTAIN, 2)) {
    pp_keyw (false, "obtain", true);
    pp_E0 (g[1]);
    pp_spc ();
    pp_C0 (g[2], large);
  }
  else if (is_func (g, MMX_LAMBDA, 2)) {
    pp_keyw (false, "lambda", true);
    pp_E0 (g[1]);
    pp_spc ();
    pp_C0 (g[2], large);
  }
  else {
    pp_E1 (g);
    pp_sep (large);
  }
}

/******************************************************************************
* Special programming constructs
******************************************************************************/

void
pretty_printer_rep::pp_Conds (generic l) {
  if (nil (l)) return;
  generic g= car (l);
  if (is_func (g, MMX_FOR, 1)) {
    pp_keyw (false, "for", true);
    pp_E0 (g[1]);
  }
  else if (is_func (g, MMX_FORIN, 2)) {
    pp_keyw (false, "for", true);
    pp_E0 (g[1]);
    pp_keyw (true, "in", true);
    pp_E0 (g[2]);
  }
  else if (is_func (g, MMX_WHILE, 1)) {
    pp_keyw (false, "while", true);
    pp_E0 (g[1]);
  }
  else if (is_func (g, MMX_UNTIL, 1)) {
    pp_keyw (false, "until", true);
    pp_E0 (g[1]);
  }
  else if (is_func (g, MMX_STEP, 1)) {
    pp_keyw (false, "step", true);
    pp_E0 (g[1]);
  }
  pp_spc ();
  pp_Conds (cdr (l));
}

void
pretty_printer_rep::pp_Splitl (generic l) {
  if (nil (l)) return;
  pp_keyw (false, "case", true);
  pp_C0 (car (l), true);
  pp_Splitl (cdr (l));
}

void
pretty_printer_rep::pp_Casel (generic l) {
  if (nil (l)) return;
  generic g= car (l);
  if (is_func (g, MMX_ELSE, 1)) {
    pp_keyw (false, "else", true);
    pp_C0 (g[1], true);
  }
  else {
    pp_Among (car (cDr (as_list (g))));
    pp_keyw (true, "do", true);
    pp_C0 (cAr (as_list (g)), true);
  }
  pp_Casel (cdr (l));
}

void
pretty_printer_rep::pp_Among (generic l) {
  if (nil (l));
  else if (nil (cdr (l)))
    pp_E0 (car (l));
  else {
    pp_T0 (car (l));
    pp_verb (",", ", ");
    pp_Among (cdr (l));
  }
}

void
pretty_printer_rep::pp_Cl (generic g) {
  if (nil (g)) return;
  pp_C0 (car (g), true);
  pp_Cl (cdr (g));
}

/******************************************************************************
* Expressions
******************************************************************************/

void
pretty_printer_rep::pp_E0 (generic g) {
  if (is_func (g, MMX_FORALL, 2)) {
    pp_verb ("\\forall\\;", "forall (");
    pp_E0 (g[1]);
    pp_verb (",", ") ");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MMX_EXISTS, 2)) {
    pp_verb ("\\exists\\;", "exists (");
    pp_E0 (g[1]);
    pp_verb (",", ") ");
    pp_E0 (g[2]);
  }
  else if (is_func (g, MMX_DEFINE, 2) ||
	   is_func (g, MMX_ASSIGN, 2) ||
	   is_func (g, MMX_DEFINEALIAS, 2) ||
	   is_func (g, MMX_PLUSASSIGN, 2) ||
	   is_func (g, MMX_MINUSASSIGN, 2) ||
	   is_func (g, MMX_TIMESASSIGN, 2) ||
	   is_func (g, MMX_OVERASSIGN, 2)) {
    pp_F0 (g[1]);
    pp_op (car (g));
    pp_E0 (g[2]);
  }
  else if (is_func (g, MMX_OBTAIN, 2)) {
    pp_keyw (false, "obtain", true);
    pp_E0 (g[1]);
    pp_spc ();
    if (is_C0 (g[2])) pp_C0 (g[2]);
    else pp_E0 (g[2]);
  }
  else if (is_func (g, MMX_LAMBDA, 2)) {
    pp_keyw (false, "lambda", true);
    pp_E0 (g[1]);
    pp_spc ();
    if (is_C0 (g[2])) pp_C0 (g[2]);
    else pp_E0 (g[2]);
  }
  else pp_E1 (g);
}

void
pretty_printer_rep::pp_E1 (generic g) {
  if (is_func (g, MMX_LESSLESS, 2) ||
      is_func (g, MMX_LESSLESS_VAR, 2) ||
      is_func (g, MMX_LESSLESS_STR, 2)) {
    pp_Lflux (g[1]);
    pp_op (car (g));
    pp_F0 (g[2]);
  }
  else if (is_func (g, MMX_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (car (g));
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

void
pretty_printer_rep::pp_Lflux (generic g) {
  if (is_func (g, MMX_LESSLESS, 2) ||
      is_func (g, MMX_LESSLESS_VAR, 2) ||
      is_func (g, MMX_LESSLESS_STR, 2)) {
    pp_Lflux (g[1]);
    pp_op (car (g));
    pp_F0 (g[2]);
  }
  else pp_F0 (g);
}

void
pretty_printer_rep::pp_Rflux (generic g) {
  if (is_func (g, MMX_GTRGTR, 2)) {
    pp_F0 (g[1]);
    pp_op (car (g));
    pp_Rflux (g[2]);
  }
  else pp_F0 (g);
}

/******************************************************************************
* Tuples and vertical tuples
******************************************************************************/

void
pretty_printer_rep::pp_E0s (generic g) {
  while (!nil (g)) {
    pp_E0 (car (g));
    g= cdr (g);
    if (nil (g)) return;
    pp_verb (",", ", ");
  }
}

void
pretty_printer_rep::pp_Vt (generic g) {
  while (!nil (g)) {
    if (is_func (car (g), MMX_TUPLE))
      pp_E0s (cdr (car (g)));
    else pp_E0 (car (g));
    g= cdr (g);
    if (nil (g)) return;
    if (texmacs_mode) out << "\\\\";
    else out << "; ";
  }
}

void
pretty_printer_rep::pp_Matrix (generic g) {
  while (!nil (g)) {
    list l (car (g));
    if (is_func (car (g), MMX_TUPLE))
      l= cdr (as_list (car (g)));
    while (!nil (l)) {
      pp_E0 (car (l));
      l= cdr (l);
      if (!nil (l)) out << "&";
    }
    g= cdr (g);
    if (nil (g)) return;
    out << "\\\\";
  }
}

void
pretty_printer_rep::pp_Vtuple (generic g) {
  bool ok = true;
  list l  = as_list (g);
  bool ht = is_func (car (l), MMX_TUPLE);
  int  len= ht? N (car (l)) - 1: 1;
  l = cdr (l);
  while (!nil (l)) {
    bool ht2 = is_func (car (l), MMX_TUPLE);
    int  len2= ht2? N (car (l)) - 1: 1;
    ok= ok && (len == len2);
    l= cdr (l);
  }
  if (ok) {
    out << "\\begin{matrix}";
    pp_Matrix (g);
    out << "\\end{matrix}";
  }
  else {
    out << "\\begin{array}{c}";
    pp_Vt (g);
    out << "\\end{array}";
  }
}

/******************************************************************************
* Formulas
******************************************************************************/

void
pretty_printer_rep::pp_F0 (generic g) {
  if (is_func (g, MMX_IMPLIES, 2) || 
      is_func (g, MMX_EQUIV, 2)) {
    pp_F1 (g[1]);
    pp_op (car (g));
    pp_F1 (g[2]);
  }
  else pp_F1 (g);
}

void
pretty_printer_rep::pp_F1 (generic g) {
  if (is_func (g, MMX_SEQOR, 2)) {
    pp_Seqors (g[1]);
    pp_keyw (true, "or", true);
    pp_Seqors (g[2]);
  }
  else if (is_func (g, MMX_OR, 2)) {
    pp_Ors (g[1]);
    pp_op (car (g));
    pp_Ors (g[2]);
  }
  else if (is_func (g, MMX_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (car (g));
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer_rep::pp_Seqors (generic g) {
  if (is_func (g, MMX_SEQOR, 2)) {
    pp_Seqors (g[1]);
    pp_keyw (true, "or", true);
    pp_Seqors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer_rep::pp_Ors (generic g) {
  if (is_func (g, MMX_OR, 2)) {
    pp_Ors (g[1]);
    pp_op (car (g));
    pp_Ors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer_rep::pp_Xors (generic g) {
  if (is_func (g, MMX_XOR, 2)) {
    pp_Xors (g[1]);
    pp_op (car (g));
    pp_Xors (g[2]);
  }
  else pp_F2 (g);
}

void
pretty_printer_rep::pp_F2 (generic g) {
  if (is_func (g, MMX_SEQAND, 2)) {
    pp_Seqands (g[1]);
    pp_keyw (true, "and", true);
    pp_Seqands (g[2]);
  }
  else if (is_func (g, MMX_AND, 2)) {
    pp_Ands (g[1]);
    pp_op (car (g));
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pretty_printer_rep::pp_Seqands (generic g) {
  if (is_func (g, MMX_SEQAND, 2)) {
    pp_Seqands (g[1]);
    pp_keyw (true, "and", true);
    pp_Seqands (g[2]);
  }
  else pp_F3 (g);
}

void
pretty_printer_rep::pp_Ands (generic g) {
  if (is_func (g, MMX_AND, 2)) {
    pp_Ands (g[1]);
    pp_op (car (g));
    pp_Ands (g[2]);
  }
  else pp_F3 (g);
}

void
pretty_printer_rep::pp_F3 (generic g) {
  if (is_func (g, MMX_EQUAL, 2) ||
      is_func (g, MMX_UNEQUAL, 2) ||
      is_func (g, MMX_LESS, 2) ||
      is_func (g, MMX_LESSEQ, 2) ||
      is_func (g, MMX_GTR, 2) ||
      is_func (g, MMX_GTREQ, 2) ||
      is_func (g, MMX_NOTLESS, 2) ||
      is_func (g, MMX_NOTLESSEQ, 2) ||
      is_func (g, MMX_NOTGTR, 2) ||
      is_func (g, MMX_NOTGTREQ, 2)) {
    pp_T0 (g[1]);
    pp_op (car (g));
    pp_T0 (g[2]);
  }
  else if (is_func (g, MMX_TYPE, 2)) {
    pp_T0 (g[1]);
    pp_verb (":", ": ");
    pp_T0 (g[2]);
  }
  else if (is_func (g, MMX_HAS, 2)) {
    pp_T0 (g[1]);
    pp_keyw (true, "has", true);
    pp_T0 (g[2]);
  }
  else if (is_func (g, MMX_HOLDS, 2)) {
    pp_T0 (g[1]);
    pp_keyw (true, "holds", true);
    pp_T0 (g[2]);
  }
  else pp_T0 (g);
}

/******************************************************************************
* Terms
******************************************************************************/

void
pretty_printer_rep::pp_T0 (generic g) {
  if (is_func (g, MMX_TRANSTYPE, 2)) {
    pp_T0 (g[1]);
    pp_op (car (g));
    pp_T1 (g[2]);
  }
  else pp_T1 (g);
}

void
pretty_printer_rep::pp_T1 (generic g) {
  if (is_func (g, MMX_FUNCTION, 2)) {
    pp_DomIm (g[1], false);
    pp_op (car (g));
    pp_DomIm (g[2], true);
  }
  else pp_T2 (g);
}

void
pretty_printer_rep::pp_T2 (generic g) {
  if (is_func (g, MMX_DOTS, 2)) {
    pp_T3 (g[1]);
    pp_op (car (g));
    pp_T3 (g[2]);
  }
  else pp_T3 (g);
}

void
pretty_printer_rep::pp_T3 (generic g) {
  if (is_func (g, MMX_PLUS, 2) ||
      is_func (g, MMX_OPLUS, 2)) {
    pp_T3 (g[1]);
    pp_op (car (g));
    pp_T3 (g[2]);
  }
  else if (is_func (g, MMX_MINUS, 2) ||
	   is_func (g, MMX_OMINUS, 2)) {
    pp_T3 (g[1]);
    pp_op (car (g));
    pp_T4 (g[2]);
  }
  else pp_T4 (g);
}

void
pretty_printer_rep::pp_T4 (generic g) {
  if (is_func (g, MMX_TIMES, 2) ||
      is_func (g, MMX_OTIMES, 2) ||
      is_func (g, MMX_COMPOSE, 2)) {
    pp_T4 (g[1]);
    pp_op (car (g));
    pp_T4 (g[2]);
  }
  else if (is_func (g, MMX_OVER, 2)) {
    if (texmacs_mode) {
      out << "\\frac{";
      pp_E0 (g[1]);
      out << "}{";
      pp_E0 (g[2]);
      out << "}";
    }
    else {
      pp_T4 (g[1]);
      pp_op (car (g));
      pp_T5 (g[2]);
    }
  }
  else if (is_func (g, MMX_OOVER, 2)) {
    pp_T4 (g[1]);
    pp_op (car (g));
    pp_T5 (g[2]);
  }
  else if (is_func (g, MMX_DIV, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "div", true);
    pp_T5 (g[2]);
  }
  else if (is_func (g, MMX_MOD, 2)) {
    pp_T4 (g[1]);
    pp_keyw (true, "mod", true);
    pp_T5 (g[2]);
  }
  else pp_T5 (g);
}

void
pretty_printer_rep::pp_T5 (generic g) {
  if (is_func (g, MMX_MINUS, 1)) {
    pp_verb ("{\\um}", "-");
    pp_R0 (g[1]);
  }
  else if (is_func (g, MMX_NOT, 1) ||
      is_func (g, MMX_PREINC, 1) ||
      is_func (g, MMX_PREDEC, 1) ||
      is_func (g, MMX_NUMBER, 1)) {
    pp_op (car (g), false);
    pp_R0 (g[1]);
  }
  else pp_R0 (g);
}

void
pretty_printer_rep::pp_DomIm (generic g, bool flag) {
  if (is_func (g, MMX_CROSS, 0)) pp_text (MMX_VOID);
  else if (is_func (g, MMX_CROSS, 1)) {
    if (flag) pp_T2 (g[1]);
    else pp_T1 (g[1]);
  }
  else if (texmacs_mode && is_func (g, MMX_CROSS)) {
    list l= cdr (as_list (g));
    while (!nil (l)) {
      pp_T5 (car (l));
      l= cdr (l);
      if (!nil (l)) out << "\\times";
    }
  }
  else if (flag) pp_T2 (g);
  else pp_T1 (g);
}

/******************************************************************************
* Radicals
******************************************************************************/

void
pretty_printer_rep::pp_R0 (generic g) {
  pp_R1 (g);
}

void
pretty_printer_rep::pp_R1 (generic g) {
  if (is_func (g, MMX_POWER, 2)) {
    if (texmacs_mode) {
      pp_R2 (g[1]);
      out << "^{";
      pp_E0 (g[2]);
      out << "}";
    }
    else {
      pp_R2 (g[1]);
      pp_op (car (g), false); // suppress spaces
      pp_R2 (g[2]);
    }
  }
  else pp_R2 (g);
}

void
pretty_printer_rep::pp_R2 (generic g) {
  if (is_func (g, MMX_POSTINC, 1) ||
      is_func (g, MMX_POSTDEC, 1) ||
      is_func (g, MMX_FACTORIAL, 1) ||
      is_func (g, MMX_PRIME, 1) ||
      is_func (g, MMX_BACKPRIME, 1)) {
    pp_R2 (g[1]);
    pp_op (car (g), false);
  }
  else if (is_func (g, MMX_DOT, 2)) {
    pp_R2 (g[1]);
    out << ".";
    pp_R3 (g[2]);
  }
  else if (is_func (g, MMX_ACCESS) && (N(g)>1)) {
    if (texmacs_mode) {
      pp_R3 (g[1]);
      out << "_{";
      pp_E0s (cdr (cdr (g)));
      out << "}";
    }
    else {
      pp_R2 (g[1]);
      out << "[";
      if ((N(g) == 3) && is_func (g[2], MMX_VTUPLE))
	pp_Vt (cdr (as_list (g[2])));
      else pp_E0s (cdr (cdr (g)));
      out << "]";
    }
  }
  else if (is_func (g, MMX_SPLIT)) {
    pp_keyw (false, "split", true);
    pp_verb ("\\left(", "(");
    pp_E0s (cdr (g));
    pp_verb ("\\right)", ")");
  }
  else if (is_func (g, MMX_JOIN, 1)) {
    pp_keyw (false, "join", true);
    pp_verb ("\\left(", "(");
    pp_E0 (g[1]);
    pp_verb ("\\right)", ")");
  }
  else if (is_func (g, MMX_SQRT, 1) && texmacs_mode) {
    out << "\\sqrt{";
    pp_E0s (cdr (g));
    out << "}";
  }
  else if (((N(g)>0) && is_applicable (g[0])) ||
	   (!nil (sl)) && (car (sl) == g)) {
    pp_R2  (car (g));
    if (!texmacs_mode) pp_spc ();
    pp_verb ("\\left(", "(");
    pp_E0s (cdr (g));
    pp_verb ("\\right)", ")");
  }
  else pp_R3 (g);
}

void
pretty_printer_rep::pp_R3 (generic g) {
  if (is_literal (g))
    pp_R3 (S (as_literal (g)));
  else if (mmx_op_type (g) != symbol ("unknown")) {
    pp_keyw (false, mmx_op_type (g), true);
    if (texmacs_mode) {
      if (g == MMX_TIMES) out << "*";
      else out << mmx_op_tmname (g);
    }
    else out << mmx_op_name (g);
  }
  else if ((gentype (g) == SYMBOL_TYPE) && (gentype (S(g)) == STRING_TYPE)) {
    string s= as_string (S(g));
    if (N(s) == 0) {
      if (texmacs_mode) out << "{\\epsilon}";
      else out << "epsilon";
    }
    else if (N(s) == 1) {
      if (texmacs_mode) out << "{" << s[0] << "}";
      else out << s[0];
    }
    else if ((s[0]>='0') && (s[0]<='9')) {
      if (texmacs_mode) {
	out << "{";
	for (int i=0; i<N(s); i++)
	  if (s[i] == 'e') {
	    int sgn= 1;
	    string r= s (i+1, N(s));
	    if (N(r) != 0 && r[0] == '+') r= r(1, N(r));
	    if (N(r) != 0 && r[0] == '-') { r= r(1, N(r)); sgn= -1; }
	    if (r == "0") break;
	    out << "}\\*{10}^{";
	    if (sgn<0) out << "{\\um}";
	    out << r;
	    break;
	  }
	  else out << s[i];
	out << "}";
      }
      else out << g;
    }
    else if (is_special_symbol (s)) {
      if (texmacs_mode)
	out << "{\\" << g << "}";
      else {
	if (starts (s, "math") && s != "math")
	  out << s (4, N(s));
	else pp_text (g);
      }
    }
    else pp_text (g);
  }
  else if (N(g) == 0)
    pp_text (g);
  else if (is_func (g, MMX_SQTUPLE)) {
    pp_verb ("\\left[", "[");
    if ((N(g) == 2) && is_func (g[1], MMX_VTUPLE)) {
      if (texmacs_mode && (N(g[1])>2))
	pp_Vtuple (cdr (as_list (g[1])));
      else pp_Vt (cdr (as_list (g[1])));
    }
    else pp_E0s (cdr (g));
    pp_verb ("\\right]", "]");
  }
  else if (is_func (g, MMX_TUPLE)) {
    pp_verb ("\\left(", "(");
    pp_E0s (cdr (g));
    pp_verb ("\\right)", ")");
  }
  else if (is_func (g, MMX_VTUPLE)) {
    pp_verb ("\\left(", "(");
    if (texmacs_mode && (N(g)>2))
      pp_Vtuple (cdr (as_list (g)));
    else pp_Vt (cdr (as_list (g)));
    pp_verb ("\\right)", ")");
  }
  else {
    pp_verb ("\\left(", "(");
    sl= cons (g, sl);
    pp_E0 (g);
    sl= cdr (sl);
    pp_verb ("\\right)", ")");
  }
}

/******************************************************************************
* Interface
******************************************************************************/

void
pretty_printer_rep::pp (generic g) {
  // cout << lf << "g= " << g << lf << lf;
  out << blank;
  if (texmacs_mode) {
    out << DATA_BEGIN << "latex:$";
    if (brown_flag) out << "{\\brown ";
  }
  if (is_func (g, MMX_BEGIN)) pp_Cl (cdr (g));
  else if (is_C0 (g)) pp_C0 (g);
  else pp_E0 (g);
  if (texmacs_mode) {
    if (brown_flag) out << "}";
    out << "$" << DATA_END;
  }
}

void
pretty_print (ostream& out, generic g) {
  pretty_printer pp (out);
  pp->pp (g);
}

} // namespace mmx
